/* Fo
 * fo-context-util.c: Utility functions for FoContext objects
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: fo-context-util.c,v 1.5 2003/07/02 11:00:26 tonygraham Exp $
 *
 * See Copying for the status of this software.
 */

#include "fo-context-util.h"
#include "fo-context-private.h"
#include "fo-property-block-progression-dimension.h"
#include "fo-property-border-after-width.h"
#include "fo-property-border-before-width.h"
#include "fo-property-border-bottom-width.h"
#include "fo-property-border-end-width.h"
#include "fo-property-border-left-width.h"
#include "fo-property-border-right-width.h"
#include "fo-property-border-start-width.h"
#include "fo-property-border-top-width.h"
#include "fo-property-end-indent.h"
#include "fo-property-height.h"
#include "fo-property-inline-progression-dimension.h"
#include "fo-property-keep-together.h"
#include "fo-property-keep-together-within-column.h"
#include "fo-property-keep-together-within-line.h"
#include "fo-property-keep-together-within-page.h"
#include "fo-property-keep-with-next.h"
#include "fo-property-keep-with-next-within-column.h"
#include "fo-property-keep-with-next-within-line.h"
#include "fo-property-keep-with-next-within-page.h"
#include "fo-property-keep-with-previous.h"
#include "fo-property-keep-with-previous-within-column.h"
#include "fo-property-keep-with-previous-within-line.h"
#include "fo-property-keep-with-previous-within-page.h"
#include "fo-property-margin-bottom.h"
#include "fo-property-margin-left.h"
#include "fo-property-margin-right.h"
#include "fo-property-margin-top.h"
#include "fo-property-padding-end.h"
#include "fo-property-padding-start.h"
#include "fo-property-start-indent.h"
#include "fo-property-space-after.h"
#include "fo-property-space-before.h"
#include "fo-property-space-end.h"
#include "fo-property-space-start.h"
#include "fo-property-width.h"
#include "fo-property-writing-mode.h"
#include "fo-area-area.h"
#include "fo-area-reference.h"

/**
 * fo_context_height_width_resolve:
 * @current_context: #FoContext of the current #FoFo.
 * @parent_context:  #FoContext of the parent #FoFo of the current #FoFo.
 * 
 * Resolve "height", "width", "inline-progression-dimension", and
 * "block-progression-dimension" property values.
 **/
void
fo_context_height_width_resolve (FoContext *current_context,
				 FoContext *parent_context)
{
  FoEnumAreaDirection bpd;
  FoProperty *writing_mode;
  FoProperty *height;
  FoProperty *width;
  FoProperty *block_progression_dimension;
  FoProperty *inline_progression_dimension;

  g_return_if_fail (FO_IS_CONTEXT (current_context));
  g_return_if_fail (FO_IS_CONTEXT (parent_context));

  height = current_context->height;
  width = current_context->width;
  block_progression_dimension =
    current_context->block_progression_dimension;
  inline_progression_dimension =
    current_context->inline_progression_dimension;

  if (current_context->writing_mode != NULL)
    writing_mode = current_context->writing_mode;
  else
    writing_mode = parent_context->writing_mode;

  bpd = fo_property_writing_mode_to_bpd (writing_mode, NULL);

  if (height == NULL &&
      width == NULL &&
      block_progression_dimension != NULL &&
      inline_progression_dimension != NULL)
    {
#if defined(LIBFO_DEBUG) && 0
      g_message ("height_width_resolve:: don't bother.");
#endif
      return;
    }

#if defined(LIBFO_DEBUG) && 0
  g_message ("height_width_resolve:: before:: bpd: %d", bpd);
  fo_object_debug_dump (writing_mode, 0);
  fo_object_debug_dump (height, 0);
  fo_object_debug_dump (width, 0);
  fo_object_debug_dump (block_progression_dimension, 0);
  fo_object_debug_dump (inline_progression_dimension, 0);
#endif

  if (block_progression_dimension == NULL)
    {
      block_progression_dimension =
	fo_property_block_progression_dimension_get_initial ();
    }

  if (inline_progression_dimension == NULL)
    {
      inline_progression_dimension =
	fo_property_inline_progression_dimension_get_initial ();
    }

  if (bpd == FO_ENUM_AREA_DIRECTION_TB ||
      bpd == FO_ENUM_AREA_DIRECTION_BT)
    {
      if (height != NULL)
	{
	  FoDatatype *height_value = fo_property_get_value (height);
	  FoDatatype *length_range = g_object_new (FO_TYPE_LENGTH_RANGE,
						   "minimum",
						   height_value,
						   "optimum",
						   height_value,
						   "maximum",
						   height_value,
						   NULL);
	  fo_property_set_value (block_progression_dimension,
				 length_range);
	}

      if (width != NULL)
	{
	  FoDatatype *width_value = fo_property_get_value (width);
	  FoDatatype *length_range = g_object_new (FO_TYPE_LENGTH_RANGE,
						   "minimum",
						   width_value,
						   "optimum",
						   width_value,
						   "maximum",
						   width_value,
						   NULL);
	  fo_property_set_value (inline_progression_dimension,
				 length_range);
	}
    }
  else if (bpd == FO_ENUM_AREA_DIRECTION_LR ||
	   bpd == FO_ENUM_AREA_DIRECTION_RL)
    {
      if (width != NULL)
	{
	  FoDatatype *width_value = fo_property_get_value (width);
	  FoDatatype *length_range = g_object_new (FO_TYPE_LENGTH_RANGE,
						   "minimum",
						   width_value,
						   "optimum",
						   width_value,
						   "maximum",
						   width_value,
						   NULL);
	  fo_property_set_value (block_progression_dimension,
				 length_range);
	}


      if (height != NULL)
	{
	  FoDatatype *height_value = fo_property_get_value (height);
	  FoDatatype *length_range = g_object_new (FO_TYPE_LENGTH_RANGE,
						   "minimum",
						   height_value,
						   "optimum",
						   height_value,
						   "maximum",
						   height_value,
						   NULL);
	  fo_property_set_value (inline_progression_dimension,
				 length_range);
	}

    }
  else
    {
      g_assert_not_reached ();
    }

  /* This may be setting the properties to the same values, but it may not,
     especially since block_progression_dimension and
     inline_progression_dimension may have been NULL initially. */
  fo_context_set_block_progression_dimension (current_context,
					      block_progression_dimension);
  fo_context_set_inline_progression_dimension (current_context,
					       inline_progression_dimension);

#if defined(LIBFO_DEBUG) && 0
  g_message ("height_width_resolve:: after:: bpd: %d", bpd);
  fo_object_debug_dump (height, 0),
    fo_object_debug_dump (width, 0),
    fo_object_debug_dump (block_progression_dimension, 0),
    fo_object_debug_dump (inline_progression_dimension, 0),
    fo_object_debug_dump (writing_mode, 0);
#endif
}

/**
 * fo_context_resolve_keeps:
 * @current_context: #FoContext of the current #FoFo.
 * @parent_context:  #FoContext of the parent #FoFo of the current #FoFo.
 * 
 * Resolve the components of the various keep properties.
 *
 * "keep-together" is inherited; "keep-with-next" and
 * "keep-with-previous" are not inherited.
 **/
void
fo_context_resolve_keeps (FoContext *current_context,
			  FoContext *parent_context)
{
  FoProperty *shortform;
  FoProperty *within_line;
  FoProperty *within_column;
  FoProperty *within_page;
  FoDatatype *use_shortform;

  shortform = fo_context_get_keep_together (current_context);
  if (shortform == NULL)
    {
      shortform = fo_context_get_keep_together (parent_context);
    }
  within_line = fo_context_get_keep_together_within_line (current_context);
  within_column = fo_context_get_keep_together_within_column (current_context);
  within_page = fo_context_get_keep_together_within_page (current_context);

  use_shortform =
    fo_keep_resolve (fo_property_get_value (shortform),
		     within_line ? fo_property_get_value (within_line) : NULL,
		     within_column ? fo_property_get_value (within_column) : NULL,
		     within_page ? fo_property_get_value (within_page) : NULL,
		     NULL);

  fo_context_set_keep_together (current_context,
				g_object_new (FO_TYPE_PROPERTY_KEEP_TOGETHER,
					      "value",
					      use_shortform,
					      NULL));
  fo_context_set_keep_together_within_line (current_context, NULL);
  fo_context_set_keep_together_within_column (current_context, NULL);
  fo_context_set_keep_together_within_page (current_context, NULL);

  shortform =
    fo_context_get_keep_with_previous (current_context);
  if (shortform == NULL)
    {
      shortform = fo_property_keep_with_previous_get_initial ();
    }
  within_line =
    fo_context_get_keep_with_previous_within_line (current_context);
  within_column =
    fo_context_get_keep_with_previous_within_column (current_context);
  within_page =
    fo_context_get_keep_with_previous_within_page (current_context);

  use_shortform =
    fo_keep_resolve (fo_property_get_value (shortform),
		     within_line ? fo_property_get_value (within_line) : NULL,
		     within_column ? fo_property_get_value (within_column) : NULL,
		     within_page ? fo_property_get_value (within_page) : NULL,
		     NULL);

  fo_context_set_keep_with_previous (current_context,
				     g_object_new (FO_TYPE_PROPERTY_KEEP_WITH_PREVIOUS,
						   "value",
						   use_shortform,
						   NULL));
  fo_context_set_keep_with_previous_within_line (current_context, NULL);
  fo_context_set_keep_with_previous_within_column (current_context, NULL);
  fo_context_set_keep_with_previous_within_page (current_context, NULL);

  shortform =
    fo_context_get_keep_with_next (current_context);
  if (shortform == NULL)
    {
      shortform = fo_property_keep_with_next_get_initial ();
    }
  within_line =
    fo_context_get_keep_with_next_within_line (current_context);
  within_column =
    fo_context_get_keep_with_next_within_column (current_context);
  within_page =
    fo_context_get_keep_with_next_within_page (current_context);

  use_shortform =
    fo_keep_resolve (fo_property_get_value (shortform),
		     within_line ? fo_property_get_value (within_line) : NULL,
		     within_column ? fo_property_get_value (within_column) : NULL,
		     within_page ? fo_property_get_value (within_page) : NULL,
		     NULL);

  fo_context_set_keep_with_next (current_context,
				 g_object_new (FO_TYPE_PROPERTY_KEEP_WITH_NEXT,
					       "value",
					       use_shortform,
					       NULL));
  fo_context_set_keep_with_next_within_line (current_context, NULL);
  fo_context_set_keep_with_next_within_column (current_context, NULL);
  fo_context_set_keep_with_next_within_page (current_context, NULL);
}

/**
 * fo_context_length_ranges_resolve:
 * @context:        The #FoContext object whose spaces will be resolved
 * @reference_area: #FoAreaReference to use when resolving percentages
 * 
 * Resolve the components of the various length-range-valued
 * properties and set @context accordingly
 **/
void
fo_context_length_ranges_resolve (FoContext *context,
				  FoArea    *reference_area)
{
  GError *tmp_error = NULL;
  FoDatatype *length_range = NULL;
  FoDatatype *shorthand, *minimum, *optimum, *maximum;
  gfloat available_width = 0.0;
  gfloat available_height = 0.0;

  g_return_if_fail (context != NULL);
  g_return_if_fail (FO_IS_CONTEXT (context));
  g_return_if_fail (reference_area == NULL ||
		    FO_IS_AREA_REFERENCE (reference_area));

#if defined(LIBFO_DEBUG) && 0
  g_message ("length_ranges_resolve:: bpdim:: shortform: %s; minimum: %s; optimum: %s; maximum: %s; reference: %s",
	     fo_object_sprintf (context->block_progression_dimension),
	     fo_object_sprintf (context->block_progression_dimension_minimum),
	     fo_object_sprintf (context->block_progression_dimension_optimum),
	     fo_object_sprintf (context->block_progression_dimension_maximum);
	     fo_object_sprintf (reference_area));
#endif

  if (reference_area != NULL)
    {
      available_width = fo_area_get_child_available_ipdim (reference_area);
      available_height = fo_area_get_child_available_bpdim (reference_area);
    }

  /* Block-progression-dimension, i.e., vertical */
  /* FIXME: won't always be vertical once can change reference orientation */
  shorthand =
    context->block_progression_dimension ?
    fo_property_get_value (context->block_progression_dimension) : NULL;
  minimum =
    context->block_progression_dimension_minimum ?
    fo_property_get_value (context->block_progression_dimension_minimum) : NULL;
  optimum =
    context->block_progression_dimension_optimum ?
    fo_property_get_value (context->block_progression_dimension_optimum) : NULL;
  maximum =
    context->block_progression_dimension_maximum ?
    fo_property_get_value (context->block_progression_dimension_maximum) : NULL;

  length_range = fo_length_range_resolve_allow_auto (shorthand,
						     minimum,
						     optimum,
						     maximum,
						     available_height,
						     &tmp_error);
#if defined(LIBFO_DEBUG) && 0
  g_message ("context_length_ranges_resolve:: bpdim::");
  fo_object_debug_dump (length_range, 0);
#endif

  if (length_range != NULL)
    {
      fo_context_set_block_progression_dimension (context,
						  g_object_new (FO_TYPE_PROPERTY_BLOCK_PROGRESSION_DIMENSION,
								"value",
								length_range,
								NULL));
    }
  else
    {
      fo_context_set_block_progression_dimension (context, NULL);
    }

  fo_context_set_block_progression_dimension_minimum (context, NULL);
  fo_context_set_block_progression_dimension_optimum (context, NULL);
  fo_context_set_block_progression_dimension_maximum (context, NULL);

  shorthand =
    context->inline_progression_dimension ?
    fo_property_get_value (context->inline_progression_dimension) : NULL;
  minimum =
    context->inline_progression_dimension_minimum ?
    fo_property_get_value (context->inline_progression_dimension_minimum) : NULL;
  optimum =
    context->inline_progression_dimension_optimum ?
    fo_property_get_value (context->inline_progression_dimension_optimum) : NULL;
  maximum =
    context->inline_progression_dimension_maximum ?
    fo_property_get_value (context->inline_progression_dimension_maximum) : NULL;

  length_range = fo_length_range_resolve_allow_auto (shorthand,
						     minimum,
						     optimum,
						     maximum,
						     available_width,
						     &tmp_error);
#if defined(LIBFO_DEBUG) && 0
  g_message ("context_length_ranges_resolve:: ipdim::");
  fo_object_debug_dump (length_range, 0);
#endif

  if (length_range != NULL)
    {
      fo_context_set_inline_progression_dimension (context,
						  g_object_new (FO_TYPE_PROPERTY_INLINE_PROGRESSION_DIMENSION,
								"value",
								length_range,
								NULL));
    }
  else
    {
      fo_context_set_inline_progression_dimension (context, NULL);
    }

  fo_context_set_inline_progression_dimension_minimum (context, NULL);
  fo_context_set_inline_progression_dimension_optimum (context, NULL);
  fo_context_set_inline_progression_dimension_maximum (context, NULL);
}

/**
 * fo_context_spaces_resolve:
 * @context:        The #FoContext object whose spaces will be resolved.
 * @reference_area: Reference area used when resolving percentages.
 * 
 * Resolve the components of the various space-valued properties and
 * set @context accordingly.
 **/
void
fo_context_spaces_resolve (FoContext *context,
			   FoArea    *reference_area)
{
  GError *tmp_error = NULL;
  FoDatatype *space;
  FoDatatype *shorthand, *minimum, *optimum, *maximum, *precedence, *condity;

  g_return_if_fail (context != NULL);
  g_return_if_fail (FO_IS_CONTEXT (context));

  shorthand =
    context->space_before ?
    fo_property_get_value (context->space_before) : NULL;
  minimum =
    context->space_before_minimum ?
    fo_property_get_value (context->space_before_minimum) : NULL;
  optimum =
    context->space_before_optimum ?
    fo_property_get_value (context->space_before_optimum) : NULL;
  maximum =
    context->space_before_maximum ?
    fo_property_get_value (context->space_before_maximum) : NULL;
  precedence =
    context->space_before_precedence ?
    fo_property_get_value (context->space_before_precedence) : NULL;
  condity =
    context->space_before_condity ?
    fo_property_get_value (context->space_before_condity) : NULL;

  space = fo_space_resolve (shorthand,
			    minimum,
			    optimum,
			    maximum,
			    precedence,
			    condity,
			    reference_area ? fo_area_get_child_available_bpdim (reference_area): 0.0,
			    &tmp_error);

  if (space != NULL)
    fo_context_set_space_before (context, g_object_new (FO_TYPE_PROPERTY_SPACE_BEFORE,
							"value",
							space,
							NULL));
  else
    fo_context_set_space_before (context, NULL);

  fo_context_set_space_before_minimum (context, NULL);
  fo_context_set_space_before_optimum (context, NULL);
  fo_context_set_space_before_maximum (context, NULL);
  fo_context_set_space_before_precedence (context, NULL);
  fo_context_set_space_before_condity (context, NULL);

  shorthand =
    context->space_after ?
    fo_property_get_value (context->space_after) : NULL;
  minimum =
    context->space_after_minimum ?
    fo_property_get_value (context->space_after_minimum) : NULL;
  optimum =
    context->space_after_optimum ?
    fo_property_get_value (context->space_after_optimum) : NULL;
  maximum =
    context->space_after_maximum ?
    fo_property_get_value (context->space_after_maximum) : NULL;
  precedence =
    context->space_after_precedence ?
    fo_property_get_value (context->space_after_precedence) : NULL;
  condity =
    context->space_after_condity ?
    fo_property_get_value (context->space_after_condity) : NULL;

  space = fo_space_resolve (shorthand,
			    minimum,
			    optimum,
			    maximum,
			    precedence,
			    condity,
			    reference_area ? fo_area_get_child_available_bpdim (reference_area): 0.0,
			    &tmp_error);

  if (space != NULL)
    fo_context_set_space_after (context, g_object_new (FO_TYPE_PROPERTY_SPACE_AFTER,
						       "value",
						       space,
						       NULL));
  else
    fo_context_set_space_after (context, NULL);

  fo_context_set_space_after_minimum (context, NULL);
  fo_context_set_space_after_optimum (context, NULL);
  fo_context_set_space_after_maximum (context, NULL);
  fo_context_set_space_after_precedence (context, NULL);
  fo_context_set_space_after_condity (context, NULL);

  shorthand =
    context->space_start ?
    fo_property_get_value (context->space_start) : NULL;
  minimum =
    context->space_start_minimum ?
    fo_property_get_value (context->space_start_minimum) : NULL;
  optimum =
    context->space_start_optimum ?
    fo_property_get_value (context->space_start_optimum) : NULL;
  maximum =
    context->space_start_maximum ?
    fo_property_get_value (context->space_start_maximum) : NULL;
  precedence =
    context->space_start_precedence ?
    fo_property_get_value (context->space_start_precedence) : NULL;
  condity =
    context->space_start_condity ?
    fo_property_get_value (context->space_start_condity) : NULL;

  space = fo_space_resolve (shorthand,
			    minimum,
			    optimum,
			    maximum,
			    precedence,
			    condity,
			    reference_area ? fo_area_get_child_available_bpdim (reference_area): 0.0,
			    &tmp_error);

  if (space != NULL)
    fo_context_set_space_start (context, g_object_new (FO_TYPE_PROPERTY_SPACE_START,
						       "value",
						       space,
						       NULL));
  else
    fo_context_set_space_start (context, NULL);

  shorthand =
    context->space_end ?
    fo_property_get_value (context->space_end) : NULL;
  minimum =
    context->space_end_minimum ?
    fo_property_get_value (context->space_end_minimum) : NULL;
  optimum =
    context->space_end_optimum ?
    fo_property_get_value (context->space_end_optimum) : NULL;
  maximum =
    context->space_end_maximum ?
    fo_property_get_value (context->space_end_maximum) : NULL;
  precedence =
    context->space_end_precedence ?
    fo_property_get_value (context->space_end_precedence) : NULL;
  condity =
    context->space_end_condity ?
    fo_property_get_value (context->space_end_condity) : NULL;

  space = fo_space_resolve (shorthand,
			    minimum,
			    optimum,
			    maximum,
			    precedence,
			    condity,
			    reference_area ? fo_area_get_child_available_ipdim (reference_area): 0.0,
			    &tmp_error);

  if (space != NULL)
    fo_context_set_space_end (context, g_object_new (FO_TYPE_PROPERTY_SPACE_END,
						     "value",
						     space,
						     NULL));
  else
    fo_context_set_space_end (context, NULL);

}

/**
 * fo_context_margin_indent_resolve:
 * @fo:              Current #FoFo.
 * @current_context: #FoContext of the current #FoFo.
 * @parent_context:  #FoContext of the parent #FoFo of the current #FoFo.
 * 
 * Resolve the 'margin-left', 'start-indent', 'margin-right', and
 * 'end-indent' property values of @fo.
 **/
void
fo_context_margin_indent_resolve (FoFo *fo,
				  FoContext *current_context,
				  FoContext *parent_context)
{
  FoProperty *current_indent_prop, *current_margin_prop;
  gfloat current_indent = 0;
  gfloat current_margin = 0;
  gfloat parent_indent, padding, border_width;

  g_return_if_fail (current_context != NULL);
  g_return_if_fail (parent_context != NULL);
  g_return_if_fail (FO_IS_CONTEXT (current_context));
  g_return_if_fail (FO_IS_CONTEXT (parent_context));

  /* FIXME: Doesn't handle case of absolute margin from shorthand
     since shorthands not supported yet. */

#if defined(LIBFO_DEBUG) && 0
  if (fo_fo_get_generate_reference_area (fo))
    {
      g_message ("margin_indent_resolve:: generates reference area: %s",
		 fo_object_debug_sprintf (fo));
    }
#endif
  

  /* left margin and start indent/padding/border-width */
  padding =
    fo_length_get_value (fo_property_get_value (fo_context_get_padding_start (current_context)));
  border_width =
    fo_length_get_value (fo_property_get_value (fo_context_get_border_start_width (current_context)));


  current_margin_prop = fo_context_get_margin_left (current_context);
  if (current_margin_prop != NULL)
    {
      current_margin =
	fo_length_get_value (fo_property_get_value (current_margin_prop));

      if (fo_fo_get_generate_reference_area (fo))
	{
	  current_indent =
	    current_margin + padding + border_width;
	}
      else
	{
	  parent_indent =
	    fo_length_get_value (fo_property_get_value (fo_context_get_start_indent (parent_context)));

	  current_indent =
	    parent_indent + current_margin + padding + border_width;
	}

      current_indent_prop =
	g_object_new (FO_TYPE_PROPERTY_START_INDENT,
		      "value",
		      g_object_new (FO_TYPE_LENGTH,
				    "value",
				    current_indent,
				    NULL),
		      NULL);
      fo_context_set_start_indent (current_context, current_indent_prop);
    }
  else /* current_margin_prop == NULL */
    {
      parent_indent =
	fo_length_get_value (fo_property_get_value (fo_context_get_start_indent (parent_context)));

      current_indent_prop = fo_context_get_start_indent (current_context);
      if (current_indent_prop == NULL)
	{
	  current_indent = parent_indent;
	}
      else
	{
	  current_indent =
	    fo_length_get_value (fo_property_get_value (current_indent_prop));
	}

      current_margin =
	current_indent - parent_indent - padding - border_width;

      current_margin_prop =
	g_object_new (FO_TYPE_PROPERTY_MARGIN_LEFT,
		      "value",
		      g_object_new (FO_TYPE_LENGTH,
				    "value",
				    current_margin,
				    NULL),
		      NULL);
      fo_context_set_margin_left (current_context, current_margin_prop);
    }
  
  /* right margin and end indent/padding/border-width */
  parent_indent =
    fo_length_get_value (fo_property_get_value (fo_context_get_end_indent (parent_context)));
  padding =
    fo_length_get_value (fo_property_get_value (fo_context_get_padding_end (current_context)));
  border_width =
    fo_length_get_value (fo_property_get_value (fo_context_get_border_end_width (current_context)));

  current_margin_prop = fo_context_get_margin_right (current_context);
  if (current_margin_prop != NULL)
    {
      current_margin =
	fo_length_get_value (fo_property_margin_right_get_value (current_margin_prop));

      if (fo_fo_get_generate_reference_area (fo))
	{
	  current_indent =
	    current_margin + padding + border_width;
	}
      else
	{
	  parent_indent =
	    fo_length_get_value (fo_property_get_value (fo_context_get_end_indent (parent_context)));

	  current_indent =
	    parent_indent + current_margin + padding + border_width;
	}

      current_indent_prop =
	g_object_new (FO_TYPE_PROPERTY_END_INDENT,
		      "value",
		      g_object_new (FO_TYPE_LENGTH,
				    "value",
				    current_indent,
				    NULL),
		      NULL);
      fo_context_set_end_indent (current_context, current_indent_prop);
    }
  else /* current_margin_prop == NULL */
    {
      parent_indent =
	fo_length_get_value (fo_property_get_value (fo_context_get_end_indent (parent_context)));

      current_indent_prop = fo_context_get_end_indent (current_context);
      if (current_indent_prop == NULL)
	{
	  current_indent = parent_indent;
	}
      else
	{
	  current_indent =
	    fo_length_get_value (fo_property_get_value (current_indent_prop));
	}

      current_margin =
	current_indent - parent_indent - padding - border_width;

      current_margin_prop =
	g_object_new (FO_TYPE_PROPERTY_MARGIN_RIGHT,
		      "value",
		      g_object_new (FO_TYPE_LENGTH,
				    "value",
				    current_margin,
				    NULL),
		      NULL);
      fo_context_set_margin_right (current_context, current_margin_prop);
    }
}

