/* Fo
 * fo-xsl-formatter.c: Boxed object type for libxslt XSLT processor
 *
 * Copyright (C) 2003 Sun Microsystems
 *
 * $Id: fo-xsl-formatter.c,v 1.7 2003/07/02 10:47:09 tonygraham Exp $
 *
 * See Copying for the status of this software.
 */

#include "libfo/fo-utils.h"
#include "libfo/fo-xml-doc-private.h"
#include "libfo/fo-doc-private.h"
#include "libfo/fo-doc-pdflib.h"
#include "libfo/fo-doc-gp.h"
#include "libfo/fo-xsl-formatter.h"
#include "fo/fo-fo.h"
#include "area/fo-area.h"
#include "result-to-fo.h"

const char *fo_xsl_formatter_error_messages [] = {
  N_("FoXslFormatter error"),
  N_("XSLT format failed"),
  N_("Unable to parse stylesheet file: %s"),
  N_("No input result tree specified"),
  N_("No output FoDoc specified"),
  N_("Specified FoDoc is not FoDoc object type"),
  N_("$SGML_CATALOG_FILES environment variable not set")
};

struct _FoXslFormatter
{
  FoXmlDoc *result_tree;

  FoDoc    *fo_doc;

  FoFo     *fo_tree;
  FoArea   *area_tree;

  guint ref_count;
};

/**
 * fo_xsl_formatter_error_quark:
 * 
 * Get the error quark for #FoXslFormatter.
 *
 * If the quark does not yet exist, create it.
 * 
 * Return value: #GQuark associated with #FoXslFormatter errors.
 **/
GQuark
fo_xsl_formatter_error_quark (void)
{
  static GQuark quark = 0;
  if (quark == 0)
    quark = g_quark_from_static_string ("FoXslFormatter error");
  return quark;
}


/**
 * fo_xsl_formatter_get_type:
 * 
 * Register the #FoXslFormatter object type.
 * 
 * Return value: #GType value of the #FoXslFormatter object type.
p **/
GType fo_xsl_formatter_get_type (void)
{
  static GType our_type = 0;
  
  if (our_type == 0)
    our_type = g_boxed_type_register_static ("FoXslFormatter",
					     (GBoxedCopyFunc) fo_xsl_formatter_ref,
					     (GBoxedFreeFunc) fo_xsl_formatter_unref);

  return our_type;
}

/**
 * fo_xsl_formatter_new:
 * 
 * Creates a new #FoXslFormatter.
 * 
 * Return value: the newly created #FoXslFormatter. Use
 * fo_xsl_formatter_unref to free the result.
 **/
FoXslFormatter *
fo_xsl_formatter_new (void)
{
  FoXslFormatter *fo_xsl_formatter = g_new0 (FoXslFormatter, 1);

  /*
  fo_xsl_formatter->fo_doc = fo_doc_pdflib_new ();
  */
  fo_xsl_formatter->ref_count = 1;

  return fo_xsl_formatter;
}

/**
 * fo_xsl_formatter_ref:
 * @fo_xsl_formatter: a #FoXslFormatter
 * 
 * Make a copy of a #FoXslFormatter.
 * 
 * Return value: a newly allocated #FoXslFormatter. This value
 *               must be freed using fo_xsl_formatter_unref().
 **/
FoXslFormatter *
fo_xsl_formatter_ref  (FoXslFormatter  *fo_xsl_formatter)
{
  g_return_val_if_fail (fo_xsl_formatter != NULL, NULL);
  g_return_val_if_fail (fo_xsl_formatter->ref_count > 0, NULL);

  fo_xsl_formatter->ref_count += 1;

  return fo_xsl_formatter;
}

/**
 * fo_xsl_formatter_unref:
 * @fo_xsl_formatter: #FoXslFormatter.
 * 
 * Unref and possibly free a #FoXslFormatter.
 **/
void
fo_xsl_formatter_unref  (FoXslFormatter  *fo_xsl_formatter)
{
  g_return_if_fail (fo_xsl_formatter != NULL);
  g_return_if_fail (fo_xsl_formatter->ref_count > 0);

  fo_xsl_formatter->ref_count -= 1;

  if (fo_xsl_formatter->ref_count == 0)
    {
      if (fo_xsl_formatter->result_tree != NULL)
	fo_xml_doc_unref (fo_xsl_formatter->result_tree);

      if (fo_xsl_formatter->fo_tree != NULL)
	g_object_unref (fo_xsl_formatter->fo_tree);

      if (fo_xsl_formatter->area_tree != NULL)
	g_object_unref (fo_xsl_formatter->area_tree);

      if (fo_xsl_formatter->fo_doc != NULL)
	g_object_unref (fo_xsl_formatter->fo_doc);

      g_free (fo_xsl_formatter);
    }
}

/**
 * fo_xsl_formatter_format:
 * @fo_xsl_formatter: #FoXslFormatter.
 * @libfo_context:    #FoLibfoContext.
 * @error:            #GError.
 * 
 * Formats @fo_xsl_formatter.
 * 
 * Return value: Success indication.
 **/
gboolean
fo_xsl_formatter_format (FoXslFormatter *fo_xsl_formatter,
			 FoLibfoContext *libfo_context,
			 GError        **error)
{
  GError *tmp_error = NULL;

  g_return_val_if_fail (fo_xsl_formatter != NULL, FALSE);
  g_return_val_if_fail (FO_IS_LIBFO_CONTEXT (libfo_context), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  if (fo_xsl_formatter->result_tree == NULL)
    {
      g_set_error (error,
		   FO_XSL_FORMATTER_ERROR,
		   FO_XSL_FORMATTER_ERROR_NO_RESULT_TREE,
		   _(fo_xsl_formatter_error_messages[FO_XSL_FORMATTER_ERROR_NO_RESULT_TREE]));
      return FALSE;
    }

  if (fo_xsl_formatter->fo_doc == NULL)
    {
      g_set_error (error,
		   FO_XSL_FORMATTER_ERROR,
		   FO_XSL_FORMATTER_ERROR_NO_FO_DOC,
		   _(fo_xsl_formatter_error_messages[FO_XSL_FORMATTER_ERROR_NO_FO_DOC]));
      return FALSE;
    }

  if (!FO_IS_DOC (fo_xsl_formatter->fo_doc))
    {
      g_set_error (error,
		   FO_XSL_FORMATTER_ERROR,
		   FO_XSL_FORMATTER_ERROR_NO_FO_DOC,
		   _(fo_xsl_formatter_error_messages[FO_XSL_FORMATTER_ERROR_NOT_FO_DOC]));
      return FALSE;
    }

  if (fo_xsl_formatter->fo_tree != NULL &&
      fo_xsl_formatter->area_tree != NULL)
    {
      return TRUE;
    }
  else
    {
      fo_xml_doc_to_fo_and_area_trees (fo_xml_doc_get_xml_doc (fo_xsl_formatter->result_tree),
				       fo_xsl_formatter->fo_doc,
				       &(fo_xsl_formatter->fo_tree),
				       &(fo_xsl_formatter->area_tree),
				       fo_libfo_context_get_debug_mode (libfo_context),
				       &tmp_error);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return FALSE;
	}
      else
	{
	  return TRUE;
	}
    }
}

/**
 * fo_xsl_formatter_get_result_tree:
 * @fo_xsl_formatter: #FoXslFormatter
 * 
 * Get the filename in @fo_xsl_formatter.
 * 
 * Return value: Filename of file parsed to make @fo_xsl_formatter.
 **/
FoXmlDoc *
fo_xsl_formatter_get_result_tree (FoXslFormatter *fo_xsl_formatter)
{
  g_return_val_if_fail (fo_xsl_formatter != NULL, NULL);

  return fo_xsl_formatter->result_tree;
}

/**
 * fo_xsl_formatter_set_result_tree:
 * @fo_xsl_formatter: #FoXslFormatter
 * @result_tree:      Result of previous parse or transformation.
 * 
 * Set the result tree in @fo_xsl_formatter.  Voids the current FO and
 * area trees.
 **/
void
fo_xsl_formatter_set_result_tree (FoXslFormatter *fo_xsl_formatter,
				  FoXmlDoc       *result_tree)
{
  g_return_if_fail (fo_xsl_formatter != NULL);

  if (fo_xsl_formatter->result_tree != NULL)
    g_free (fo_xsl_formatter->result_tree);

  if (fo_xsl_formatter->fo_tree != NULL)
    {
      g_object_unref (fo_xsl_formatter->fo_tree);
      fo_xsl_formatter->fo_tree = NULL;
    }

  if (fo_xsl_formatter->area_tree != NULL)
    {
      g_object_unref (fo_xsl_formatter->area_tree);
      fo_xsl_formatter->area_tree = NULL;
    }

  fo_xsl_formatter->result_tree = fo_xml_doc_ref (result_tree);
}

/**
 * fo_xsl_formatter_set_fo_doc:
 * @fo_xsl_formatter: #FoXslFormatter
 * @fo_doc:           #FoDoc for use as output of formatter.
 * 
 * Set the #FoDoc in @fo_xsl_formatter.  Voids the current FO and area
 * trees.
 **/
void
fo_xsl_formatter_set_fo_doc (FoXslFormatter *fo_xsl_formatter,
			     FoDoc          *fo_doc)
{
  g_return_if_fail (fo_xsl_formatter != NULL);

  if (fo_xsl_formatter->fo_doc != NULL)
    g_object_unref (fo_xsl_formatter->fo_doc);

  if (fo_xsl_formatter->fo_tree != NULL)
    {
      g_object_unref (fo_xsl_formatter->fo_tree);
      fo_xsl_formatter->fo_tree = NULL;
    }

  if (fo_xsl_formatter->area_tree != NULL)
    {
      g_object_unref (fo_xsl_formatter->area_tree);
      fo_xsl_formatter->area_tree = NULL;
    }

  fo_xsl_formatter->fo_doc = g_object_ref (fo_doc);
}

/**
 * fo_xsl_formatter_get_fo_tree:
 * @fo_xsl_formatter: #FoXslFormatter
 * 
 * Get the #FoFo tree created by @fo_xsl_formatter.
 * 
 * Return value: #FoFo tree created by @fo_xsl_formatter.
 **/
GObject *
fo_xsl_formatter_get_fo_tree (FoXslFormatter *fo_xsl_formatter)
{
  g_return_val_if_fail (fo_xsl_formatter != NULL, NULL);

  return G_OBJECT (fo_xsl_formatter->fo_tree);
}

/**
 * fo_xsl_formatter_get_area_tree:
 * @fo_xsl_formatter: #FoXslFormatter
 * 
 * Get the #FoArea tree created by @fo_xsl_formatter.
 * 
 * Return value: #FoArea tree created by @fo_xsl_formatter.
 **/
GObject *
fo_xsl_formatter_get_area_tree (FoXslFormatter *fo_xsl_formatter)
{
  g_return_val_if_fail (fo_xsl_formatter != NULL, NULL);

  return G_OBJECT (fo_xsl_formatter->area_tree);
}
